//////////////////////////////////////////////////////////////////////////
//
// Das Labor des Dr. B. Knackt
// 
// Copyright 2003 by Oliver Rendelmann
//
// Dieses Projekt ist ausschlielich zu Lernzwecken entstanden
//
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// INCLUDES
//////////////////////////////////////////////////////////////////////////

#include <mylua.h>
#include <iostream>
#include <fstream>
#include <list>
#include <algorithm>

#include <windows.h>
#include <conio.h>

using namespace std;

//////////////////////////////////////////////////////////////////////////
// DEFINES
//////////////////////////////////////////////////////////////////////////

#define UP		0
#define RIGHT	1
#define DOWN	2
#define LEFT	3

#define UPARROW		72
#define RIGHTARROW	77
#define DOWNARROW	80
#define LEFTARROW	75
#define ESCAPE		27

//////////////////////////////////////////////////////////////////////////
// CONSOLE HELPERS
//////////////////////////////////////////////////////////////////////////

void wait(DWORD ms)
{
	DWORD now = timeGetTime();

	while((timeGetTime() - now) < ms);
}

bool gotoxy(short x, short y)
{
     HANDLE aConsoleHandle = GetStdHandle(STD_OUTPUT_HANDLE);

     if(INVALID_HANDLE_VALUE == aConsoleHandle)
          return false;

     COORD aCoords;
     aCoords.X = x;
     aCoords.Y = y;

     SetConsoleCursorPosition(aConsoleHandle, aCoords);

	 return true;
}

int readkey()
{
	if(!_kbhit())
		return 0;

	int foo = _getch();

	if(foo == 224)
		foo = _getch();

	return foo;
}

// HACKHACK: Is kinda ugly...
void clrscr()
{
	gotoxy(0, 0);

	for(int i = 0; i < 2800; ++i)
		cout << " ";

	gotoxy(0, 0);
}


//////////////////////////////////////////////////////////////////////////
// STRUCTS
//////////////////////////////////////////////////////////////////////////

struct ent
{
	int id;			// Entity's ID... this has to be unique
	int x;			// Position
	int y;
	bool locked;	// Is the entity locked?
	char face;		// Appearance on screen
};

//////////////////////////////////////////////////////////////////////////
// FUNCTORS
//////////////////////////////////////////////////////////////////////////

// This f'tor prints an entity on screen
template<typename T>
struct show_ent : public unary_function<T, bool>
{
	void operator()(T * p)
	{
		gotoxy(p->x, p->y);
		cout << p->face;
	}
};

// This f'tor checks if an entity has a certain ID
template<typename T>
struct has_id : public unary_function<T, bool>
{
	// Constructor
	has_id(int id) : _id(id) {}

	bool operator()(T * p)
	{
		return (p->id == _id);
	}

	int _id;
};

// This f'tor deletes an object
template<typename T>
struct delete_ : public unary_function<T, bool>
{
   void operator()(T * a)
   {
      delete a;
   }
}; 

//////////////////////////////////////////////////////////////////////////
// GLOBALS
//////////////////////////////////////////////////////////////////////////

char map[20][20];

ent * player;
list<ent *> entities;

string message;

bool running = true;
bool redraw = true;

lua_State * L;

//////////////////////////////////////////////////////////////////////////
// PROTOTYPES
//////////////////////////////////////////////////////////////////////////

void Intro();
ent * CreateEntity(int id, int x, int y, char face);
void DeleteEntity(int id);
void LoadMap();
void ShowMap();
void MoveEntity(ent * p, int dir);
ent * FindByID(int id);

//////////////////////////////////////////////////////////////////////////
// LUA GLUE FUNCTIONS
//////////////////////////////////////////////////////////////////////////

static int lua_ShowMessage(lua_State * L)
{
	// HACK: Delete the old stuff
	gotoxy(0, 20);
	for(int i = 0; i < message.length(); ++i)
		cout << ' ';

	message = lua_tostring(L, -1);

	gotoxy(0, 20);
	cout << message;

	redraw = true;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_PutTile(lua_State * L)
{
	int x = lua_tonumber(L, -3);
	int y = lua_tonumber(L, -2);
	int c = lua_tonumber(L, -1);

	map[x][y] = (char)c;

	redraw = true;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_Spawn(lua_State * L)
{
	int id   = lua_tonumber(L, -4);
	int x    = lua_tonumber(L, -3);
	int y    = lua_tonumber(L, -2);
	int face = lua_tonumber(L, -1);

	ent * tmp = CreateEntity(id, x, y, (char)face);
	entities.push_back(tmp);

	redraw = true;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_Move(lua_State * L)
{
	int id  = lua_tonumber(L, -2);
	int dir = lua_tonumber(L, -1);

	MoveEntity(FindByID(id), dir);

	redraw = true;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_Lock(lua_State * L)
{
	int  id   = lua_tonumber(L, -2);
	bool lock = lua_toboolean(L, -1);

	FindByID(id)->locked = lock;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_Kill(lua_State * L)
{
	int id = lua_tonumber(L, -1);

	DeleteEntity(id);

	redraw = true;

	return 0;
}

//////////////////////////////////////////////////////////////////////////

static int lua_GetPos(lua_State * L)
{
	int id = lua_tonumber(L, -1);

	ent * tmp = FindByID(id);

	lua_pushnumber(L, tmp->x);
	lua_pushnumber(L, tmp->y);

	return 2;
}

//////////////////////////////////////////////////////////////////////////
// REGISTER FUNCTION
//////////////////////////////////////////////////////////////////////////

void RegisterGlueFunctions(lua_State * L)
{
	lua_register(L, "showmessage", lua_ShowMessage);
	lua_register(L, "puttile", lua_PutTile);
	lua_register(L, "spawn", lua_Spawn);
	lua_register(L, "move", lua_Move);
	lua_register(L, "lock", lua_Lock);
	lua_register(L, "kill", lua_Kill);
	lua_register(L, "getpos", lua_GetPos);
}

//////////////////////////////////////////////////////////////////////////
// MAIN()
//////////////////////////////////////////////////////////////////////////

void main()
{
	// Init game
	L = lua_open();						// Start up Lua
	lua_baselibopen(L);					// Open Lua's base library
	RegisterGlueFunctions(L);			// Register functions with Lua
	lua_dofile(L, "script.txt");		// Execute the script
	LoadMap();							// Load the level map
	player = CreateEntity(666, 1, 1, 2);// ... and finally create the player object

	Intro();

	while(running)
	{
		int key = readkey();

		switch(key)
		{
		case UPARROW:		MoveEntity(player, UP);		break;
		case RIGHTARROW:	MoveEntity(player, RIGHT);	break;
		case DOWNARROW:		MoveEntity(player, DOWN);	break;
		case LEFTARROW:		MoveEntity(player, LEFT);	break;
		case ESCAPE:		running = false;			break;
		}

		if(redraw)
		{
			ShowMap();
			for_each(entities.begin(), entities.end(), show_ent<ent>());
			
			gotoxy(player->x, player->y);
			cout << player->face;

			gotoxy(0, 0);

			redraw = false;
		}

		// Return back to script after yielding
		lua_dostring(L, "coroutine.resume(co)");

		// Add a small delay
		wait(30);
	}

	// Delete all remaining entites
	for_each(entities.begin(), entities.end(), delete_<ent>());
	lua_close(L);
}

//////////////////////////////////////////////////////////////////////////
// FUNCTIONS
//////////////////////////////////////////////////////////////////////////

void LoadMap()
{
	char c;
	int x = 0, y = 0;
	ifstream in("map.txt");

	while(!in.eof())
	{
		in >> c;
		map[x][y] = c;

		++x;

		if(x == 20)
		{
			x = 0;
			++y;
		}
	}

	in.close();
}

//////////////////////////////////////////////////////////////////////////

// http://www.asciitable.com/ !!

#define PUT(x, y, c)gotoxy((x),(y));cout << ((char)c)
void ShowMap()
{
	for(int y = 0; y < 20; ++y)
	{
		for(int x = 0; x < 20; ++x)
		{
			switch(map[x][y])						
			{
			case '0' : PUT(x, y, ' '); break;
			case '1' : PUT(x, y, 178); break;
				
			default : PUT(x, y, 254); break;
			}
		}
	}
}
#undef PUT

//////////////////////////////////////////////////////////////////////////

ent * CreateEntity(int id, int x, int y, char face)
{
	ent * tmp = new ent;

	tmp->id = id;
	tmp->x = x;
	tmp->y = y;
	tmp->face = face;
	tmp->locked = false;

	redraw = true;

	return tmp;
}

//////////////////////////////////////////////////////////////////////////

void DeleteEntity(int id)
{
	entities.remove(FindByID(id));

	redraw = true;
}

//////////////////////////////////////////////////////////////////////////

void MoveEntity(ent * p, int dir)
{
	// Movement not allowed if the entity is locked
	if(p->locked)
		return;

	redraw = true;

	int dx = p->x, dy = p->y;
	switch(dir)
	{
	case 0: dy--; break;
	case 1: dx++; break;
	case 2: dy++; break;
	case 3: dx--; break;
	}

	// Don't move if there is a wall
	if(map[dx][dy] != '1')
	{
		p->x = dx;
		p->y = dy;
	
		// If the field is not empty space then it has to be a trigger, so execute it
		if(map[dx][dy] != '0')
		{
			string command = "co = coroutine.create(";
			command += map[dx][dy];
			command += ")";

			lua_dostring(L, command.c_str());
		}
	}
}

//////////////////////////////////////////////////////////////////////////

ent * FindByID(int id)
{
	ent * tmp;

	if(id == 666)		// Player always is 666
		tmp = player;
	else				// Otherwise, look for it... thank God for the STL :)
		tmp = *(find_if(entities.begin(), entities.end(), has_id<ent>(id)));

	// Remark: find_if returns an iterator... by dereferencing it you get the object itself

	return tmp;
}

//////////////////////////////////////////////////////////////////////////

void Intro()
{
	clrscr();

	cout << " " << endl
		 << " *******" << endl
		 << " * Das Labor des Dr. B. Knackt * " << endl
		 << "                             *** " << endl
		 << " " << endl
		 << " Ein Spiel von Oliver Rendelmann " << endl
		 << " " << endl
		 << " Story:" << endl
		 << " Du hast von einer sehr streng geheimen Organisation" << endl
		 << " den seeehr wichtigen Auftrag bekommen das Labor des" << endl
		 << " berchtigten Dr. B. Knackt zu infiltrieren und seine" << endl
		 << " neueste Kreation zu stehlen... dieser Wahnsinnige" << endl
		 << " hat eine saumaechtige Biowaffe entwickelt. Das Projekt" << endl
		 << " laeuft unter dem Codenamen \"Experiment 626\"..." << endl
		 << " nu beweg endlich deinen faulen Arsch und hol das Vieh!!!" << endl
		 << " Amen." << endl;

	while(!kbhit());

	clrscr();
}
